"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper

from dune.perftool.generation import (c_instruction,
                                      domain,
                                      expr_instruction,
                                      globalarg,
                                      iname,
                                      temporary_variable,
                                      valuearg,
                                      )

from ufl.algorithms import MultiFunction

import loopy


@iname
def dimension_iname(index):
    from dune.perftool.pdelab import name_index
    from dune.perftool.pdelab.geometry import name_dimension
    iname = name_index(index)
    dimname = name_dimension()
    domain(iname, dimname)
    return iname


@iname
def argument_iname(arg):
    # TODO extract the {iname}_n thing by a preamble
    from dune.perftool.ufl.modified_terminals import modified_argument_number
    iname = "arg{}".format(chr(ord("i") + modified_argument_number()(arg)))
    domain(iname, iname + "_n")
    return iname


@iname
def quadrature_iname():
    domain("q", "q_n")
    return "q"


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper):
    def __init__(self):
        super(UFL2LoopyVisitor, self).__init__()

    def coefficient(self, o):
        # All trial functions should already be handled by the TrialFunctionExtractor class!
        assert o.count() != 0
        # TODO implement non-trialfunction coefficients
        raise NotImplementedError

    # TODO use multiple inheritance and have a geometry transformer in the pdelab subpackage
    def facet_area(self, o):
        from pymbolic.primitives import Variable
        from dune.perftool.pdelab.geometry import name_facetarea
        return Variable(name_facetarea())


def get_pymbolic_expr(expr):
    """ Transform the given UFL expression into a pymbolic expression
    and have all sorts of side effects on the generation cache. """
    # We do need to manually handle modified terminals related to trial functions
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    from dune.perftool.ufl.transformations.replace import ReplaceExpression
    from dune.perftool.pdelab.argument import name_trialfunction
    from pymbolic.primitives import Variable

    trial_ma = extract_modified_arguments(expr, trialfunction=True)
    # OLD CODE had: globalarg(name)
    rmap = {ma: Variable(name_trialfunction(ma)) for ma in trial_ma}
    ufl2l_mf = UFL2LoopyVisitor()
    re_mf = ReplaceExpression(replacemap=rmap, otherwise=ufl2l_mf)
    ufl2l_mf.call = re_mf.__call__

    return re_mf(expr)


class _Counter:
    counter = 0


def get_count():
    c = _Counter.counter
    _Counter.counter = c + 1
    return c


def transform_accumulation_term(term):
    # Get the accumulation expression and the modified arguments
    expr, args = term

    # We always have a quadrature loop
    quadrature_iname()

    # Get the pymbolic expression needed for this accumulation term.
    # This includes filling the cache with all sorts of necessary preambles!
    pymbolic_expr = get_pymbolic_expr(expr)

    # Now simplify the expression
    # TODO: Add a switch to disable/configure this.
    from dune.perftool.pymbolic.simplify import simplify_pymbolic_expression
    pymbolic_expr = simplify_pymbolic_expression(pymbolic_expr)

    # Define a temporary variable for this expression
    expr_tv_name = "expr_" + str(get_count()).zfill(4)
    expr_tv = temporary_variable(expr_tv_name)
    from pymbolic.primitives import Variable
    expr_instruction(loopy.ExpressionInstruction(assignee=Variable(expr_tv_name), expression=pymbolic_expr))

    # The data that is used to collect the arguments for the accumulate function
    accumargs = []
    argument_code = []

    # Generate the code for the modified arguments:
    for arg in args:
        from dune.perftool.pdelab.argument import name_argumentspace, name_argument
        accumargs.append(name_argumentspace(arg))
        accumargs.append(argument_iname(arg))
        name = name_argument(arg)
        argument_code.append(name)
        globalarg(name)

    from dune.perftool.pdelab.argument import name_residual
    residual = name_residual()

    from dune.perftool.generation import retrieve_cache_items
    inames = retrieve_cache_items("iname")

    from dune.perftool.pdelab.quadrature import name_factor
    c_instruction(loopy.CInstruction(inames,
                                     "{}.accumulate({}, {}*{}*{})".format(residual,
                                                                          ", ".join(accumargs),
                                                                          expr_tv_name,
                                                                          "*".join(argument_code),
                                                                          name_factor()
                                                                          )
                                     )
                  )
